allowed_alphabet = "abcdefghijklmnopqrstuvwxyz1234567890"
allowed_alphabet += allowed_alphabet.upper()


def name(given):
    """
    Clean a given name so that it can be used inside of JCI.
    """
    return "".join(l if l in allowed_alphabet else "-" for l in given)



"""
A docker executor for Jaypore CI.
"""
import pendulum
import docker
from rich import print as rprint

from jaypore_ci import clean
from jaypore_ci.interfaces import Executor, TriggerFailed, JobStatus
from jaypore_ci.logging import logger


class Docker(Executor):
    """
    Run jobs via docker.

    This will:
        - Create a separate network for each run
        - Run jobs as part of the network
        - Clean up all jobs when the pipeline exits.
    """

    def __init__(self):
        super().__init__()
        self.pipe_id = None
        self.pipeline = None
        self.docker = docker.from_env()
        self.client = docker.APIClient()
        self.__execution_order__ = []

    def logging(self):
        """
        Returns a logging instance that has executor specific
        information bound to it.
        """
        return logger.bind(pipe_id=self.pipe_id, network_name=self.get_net())

    def set_pipeline(self, pipeline):
        """
        Set executor's pipeline to the given one.

        This will clean up old networks and create new ones.
        """
        if self.pipe_id is not None:
            self.delete_network()
            self.delete_all_jobs()
        self.pipe_id = pipeline.pipe_id
        self.pipeline = pipeline
        self.create_network()

    def __exit__(self, exc_type, exc_value, traceback):
        self.delete_network()
        self.delete_all_jobs()

    def get_net(self):
        """
        Return a network name based on what the curent pipeline is.
        """
        return f"jayporeci__net__{self.pipe_id}" if self.pipe_id is not None else None

    def create_network(self):
        """
        Will create a docker network.

        If it fails to do so in 3 attempts it will raise an
        exception and fail.
        """
        assert self.pipe_id is not None, "Cannot create network if pipe is not set"
        for _ in range(3):
            if len(self.docker.networks.list(names=[self.get_net()])) != 0:
                self.logging().info("Found network", network_name=self.get_net())
                return
            self.logging().info(
                "Create network",
                subprocess=self.docker.networks.create(
                    name=self.get_net(), driver="bridge"
                ),
            )
        raise TriggerFailed("Cannot create network")

    def delete_all_jobs(self):
        """
        Deletes all jobs associated with the pipeline for this
        executor.

        It will stop any jobs that are still running.
        """
        assert self.pipe_id is not None, "Cannot delete jobs if pipe is not set"
        job = None
        for job in self.pipeline.jobs.values():
            if job.run_id is not None and not job.run_id.startswith("pyrun_"):
                container = self.docker.containers.get(job.run_id)
                container.stop(timeout=1)
                self.logging().info("Stop job:", run_id=job.run_id)
                job.check_job(with_update_report=False)
        if job is not None:
            job.check_job()
        self.logging().info("All jobs stopped")

    def delete_network(self):
        """
        Delete the network for this executor.
        """
        assert self.pipe_id is not None, "Cannot delete network if pipe is not set"
        try:
            net = self.docker.networks.get(self.get_net())
            net.remove()
        except docker.errors.NotFound:
            self.logging().error("Delete network: Not found", netid=self.get_net())

    def get_job_name(self, job, tail=False):
        """
        Generates a clean job name slug.
        """
        name = clean.name(job.name)
        if tail:
            return name
        return f"jayporeci__job__{self.pipe_id}__{name}"

    def run(self, job: "Job") -> str:
        """
        Run the given job and return a docker container ID.
        In case something goes wrong it will raise TriggerFailed
        """
        assert self.pipe_id is not None, "Cannot run job if pipe id is not set"
        trigger = {
            "detach": True,
            "environment": job.get_env(),
            "volumes": [
                "/var/run/docker.sock:/var/run/docker.sock",
                "/usr/bin/docker:/usr/bin/docker:ro",
                f"/tmp/jayporeci__src__{self.pipeline.remote.sha}:/jaypore_ci/run",
            ],
            "name": self.get_job_name(job),
            "network": self.get_net(),
            "image": job.image,
            "command": job.command if not job.is_service else None,
        }
        if not job.is_service:
            trigger["working_dir"] = "/jaypore_ci/run"
        if not job.is_service:
            assert job.command
        rprint(trigger)
        try:
            container = self.docker.containers.run(**trigger)
            self.__execution_order__.append(
                (self.get_job_name(job, tail=True), container.id, "Run")
            )
            return container.id
        except docker.errors.APIError as e:
            self.logging().exception(e)
            raise TriggerFailed(e) from e

    def get_status(self, run_id: str) -> JobStatus:
        """
        Given a run_id, it will get the status for that run.
        """
        inspect = self.client.inspect_container(run_id)
        status = JobStatus(
            is_running=inspect["State"]["Running"],
            exit_code=int(inspect["State"]["ExitCode"]),
            logs="",
            started_at=pendulum.parse(inspect["State"]["StartedAt"]),
            finished_at=pendulum.parse(inspect["State"]["FinishedAt"])
            if inspect["State"]["FinishedAt"] != "0001-01-01T00:00:00Z"
            else None,
        )
        # --- logs
        self.logging().debug("Check status", status=status)
        logs = self.docker.containers.get(run_id).logs().decode()
        return status._replace(logs=logs)

    def get_execution_order(self):
        return {name: i for i, (name, *_) in enumerate(self.__execution_order__)}



import re

ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def clean_logs(logs):
    """
    Clean logs so that they don't have HTML/ANSI color codes in them.
    """
    for old, new in [("<", r"\<"), (">", r"\>"), ("`", '"'), ("\r", "\n")]:
        logs = logs.replace(old, new)
    return [line.strip() for line in ansi_escape.sub("", logs).split("\n")]



import pendulum
from jaypore_ci.interfaces import Reporter, Status


def __get_time_format__(job):
    time = " --:--"
    if job.run_state is not None:
        if (
            job.run_state.finished_at is not None
            and job.run_state.started_at is not None
        ):
            s = job.run_state.finished_at - job.run_state.started_at
        elif job.run_state.started_at is not None:
            s = pendulum.now() - job.run_state.started_at
        else:
            s = None
        s = s.in_seconds() if s is not None else 0
        m = s // 60
        time = f"{m:>3}:{s % 60:>2}"
    return time


def get_job_report(jobname):
    with open(f"/jaypore_ci/run/{jobname}.txt", "r", encoding="utf-8") as fl:
        return fl.read()


__ST_MAP__ = {
    Status.RUNNING: "🔵",
    Status.FAILED: "🔴",
    Status.PASSED: "🟢",
}


class Text(Reporter):
    def render(self, pipeline):
        """
        Returns a human readable report for a given pipeline.
        """
        max_name = max(len(job.name) for job in pipeline.jobs.values())
        max_name = max(max_name, len("jayporeci"))
        max_report = 10
        name = ("JayporeCI" + " " * max_name)[:max_name]
        graph = [
            "",
            "```jayporeci",
            f"╔ {pipeline.get_status_dot()} : {name} [sha {pipeline.remote.sha[:10]}]",
        ]
        closer = "┗" + ("━" * (len(" O : ") + max_name + 1 + 1 + 8 + 1)) + "┛"
        for stage in pipeline.stages:
            nodes, edges = set(), set()
            for job in pipeline.jobs.values():
                if job.stage != stage:
                    continue
                nodes.add(job.name)
                edges |= {(p, job.name) for p in job.parents}
            if not nodes:
                continue
            graph += [f"┏━ {stage}", "┃"]
            for n in sorted(
                nodes, key=lambda x: (len(pipeline.jobs[x].parents), x)
            ):  # Fewer parents first
                n = pipeline.jobs[n]
                name = (n.name + " " * max_name)[:max_name]
                status = __ST_MAP__.get(n.status, "🟡")
                run_id = f"{n.run_id}"[:8] if n.run_id is not None else ""
                graph += [f"┃ {status} : {name} [{run_id:<8}] {__get_time_format__(n)}"]
                try:
                    report = get_job_report(n.name)
                    report = " ".join(report.strip().split())
                    report = (report + " " * max_report)[:max_report]
                except FileNotFoundError:
                    report = " " * max_report
                graph[-1] += f" {report}"
                if n.parents:
                    graph[-1] += f" ❮-- {n.parents}"
            graph += [closer]
        graph += ["```"]
        graph = "\n".join(graph)
        return f"\n{graph}"



"""
A github remote git host.

This is used to report pipeline status to the remote.
"""
import os
from pathlib import Path
from urllib.parse import urlparse

import requests

from jaypore_ci.interfaces import Remote, RemoteApiFailed, Repo
from jaypore_ci.logging import logger


class Github(Remote):  # pylint: disable=too-many-instance-attributes
    """
    The remote implementation for github.
    """

    def __headers__(self):
        return {
            "Authorization": f"Bearer {self.token}",
            "Accept": "application/vnd.github+json",
            "X-Github-Api-Version": "2022-11-28",
        }

    @classmethod
    def from_env(cls, *, repo: Repo) -> "Github":
        """
        Creates a remote instance from the environment.
        It will:

            - Find the remote location using `git remote`.
            - Find the current branch
            - Create a new pull request for that branch
            - Allow posting updates using the gitea token provided
        """
        remote = urlparse(repo.remote)
        os.environ["JAYPORE_COMMIT_BRANCH"] = repo.branch
        os.environ["JAYPORE_COMMIT_SHA"] = repo.sha
        return cls(
            root="https://api.github.com",
            owner=Path(remote.path).parts[1],
            repo=Path(remote.path).parts[2].replace(".git", ""),
            branch=repo.branch,
            token=os.environ["JAYPORE_GITHUB_TOKEN"],
            sha=repo.sha,
        )

    def __init__(
        self, *, root, owner, repo, token, **kwargs
    ):  # pylint: disable=too-many-arguments
        super().__init__(**kwargs)
        # --- customer
        self.root = root
        self.api = root
        self.owner = owner
        self.repo = repo
        self.token = token
        self.timeout = 10
        self.base_branch = "main"

    def logging(self):
        """
        Return's a logging instance with information about gitea bound to it.
        """
        return logger.bind(
            root=self.root, owner=self.owner, repo=self.repo, branch=self.branch
        )

    def get_pr_id(self):
        """
        Returns the pull request ID for the current branch.
        """
        r = requests.post(
            f"{self.api}/repos/{self.owner}/{self.repo}/pulls",
            headers=self.__headers__(),
            timeout=self.timeout,
            json={
                "base": self.base_branch,
                "body": "Branch auto created by JayporeCI",
                "head": self.branch,
                "title": self.branch,
            },
        )
        self.logging().debug("Create PR", status_code=r.status_code)
        if r.status_code == 201:
            return r.json()["number"]
        r = requests.get(
            f"{self.api}/repos/{self.owner}/{self.repo}/pulls",
            headers=self.__headers__(),
            timeout=self.timeout,
            json={"base": self.base_branch, "head": self.branch, "draft": True},
        )
        self.logging().debug("Get PR", status_code=r.status_code)
        if r.status_code == 200:
            if len(r.json()) == 1:
                return r.json()[0]["number"]
        self.logging().debug(
            "Failed github api",
            api=self.api,
            owner=self.owner,
            repo=self.repo,
            token=self.token,
            branch=self.branch,
            status=r.status_code,
            response=r.text,
        )
        raise RemoteApiFailed(r)

    def publish(self, report: str, status: str):
        """
        Will publish the report to the remote.

        :param report: Report to write to remote.
        :param status: One of ["pending", "success", "error", "failure"]
                       This is the dot/tick next to each commit in gitea.
        """
        assert status in ("pending", "success", "error", "failure")
        issue_id = self.get_pr_id()
        # Get existing PR body
        r = requests.get(
            f"{self.api}/repos/{self.owner}/{self.repo}/pulls/{issue_id}",
            timeout=self.timeout,
            headers=self.__headers__(),
        )
        self.logging().debug("Get existing body", status_code=r.status_code)
        assert r.status_code == 200
        body = r.json()["body"]
        body = (line for line in body.split("\n"))
        prefix = []
        for line in body:
            if "```jayporeci" in line:
                prefix = prefix[:-1]
                break
            prefix.append(line)
        while prefix and prefix[-1].strip() == "":
            prefix = prefix[:-1]
        prefix.append("")
        # Post new body with report
        report = "\n".join(prefix) + "\n" + report
        r = requests.patch(
            f"{self.api}/repos/{self.owner}/{self.repo}/pulls/{issue_id}",
            json={"body": report},
            timeout=self.timeout,
            headers=self.__headers__(),
        )
        self.logging().debug("Published new report", status_code=r.status_code)
        # Set commit status
        r = requests.post(
            f"{self.api}/repos/{self.owner}/{self.repo}/statuses/{self.sha}",
            json={
                "context": "JayporeCi",
                "description": f"Pipeline status is: {status}",
                "state": status,
            },
            timeout=self.timeout,
            headers=self.__headers__(),
        )
        self.logging().debug(
            "Published new status", status=status, status_code=r.status_code
        )



"""
This is used to save the pipeline status to git itself.
"""
import time
import subprocess

from jaypore_ci.interfaces import Remote
from jaypore_ci.repos import Git
from jaypore_ci.logging import logger


class GitRemote(Remote):  # pylint: disable=too-many-instance-attributes
    """
    You can save pipeline status to git using this remote.

    To push/fetch your local refs to a git remote you can run

    .. code-block:: console

        git fetch   origin refs/jayporeci/*:refs/jayporeci/*
        git push    origin refs/jayporeci/*:refs/jayporeci/*
    """

    @classmethod
    def from_env(cls, *, repo: Git) -> "GitRemote":
        """
        Creates a remote instance from the environment.
        """
        assert isinstance(repo, Git), "Git remote can only work in a git repo"
        return cls(
            repo=repo,
            branch=repo.branch,
            sha=repo.sha,
        )

    def __init__(self, *, repo, **kwargs):
        super().__init__(**kwargs)
        self.repo = repo

    def logging(self):
        """
        Return's a logging instance with information about git bound to it.
        """
        return logger.bind(repo=self.repo)

    def publish(self, report: str, status: str) -> None:
        """
        Will publish the report via email.

        :param report: Report to write to remote.
        :param status: One of ["pending", "success", "error", "failure",
            "warning"] This is the dot next to each commit in gitea.
        """
        assert status in ("pending", "success", "error", "failure", "warning")
        now = time.time()
        lines = ""
        git_blob_sha = subprocess.check_output(
            "git hash-object -w --stdin",
            input=report,
            text=True,
            stderr=subprocess.STDOUT,
            shell=True,
        ).strip()
        lines += f"\n100644 blob {git_blob_sha}\t{now}.txt"
        lines = lines.strip()
        git_tree_sha = subprocess.run(
            "git mktree",
            input=lines,
            text=True,
            shell=True,
            check=False,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        ).stdout.strip()
        git_commit_sha = subprocess.run(
            f"git commit-tree {git_tree_sha}",
            text=True,
            input=f"JayporeCI status: {now}",
            shell=True,
            check=False,
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
        )
        assert git_commit_sha.returncode == 0
        git_commit_sha = (
            subprocess.check_output(
                f"git update-ref refs/jayporeci/{self.repo.sha} {git_commit_sha.stdout.strip()}",
                shell=True,
                stderr=subprocess.STDOUT,
            )
            .decode()
            .strip()
        )
        self.logging().info(
            "Published status to local git: refs/jayporeci/{self.repo.sha} {git_commit_sha}"
        )



"""
Defines interfaces for remotes and executors.

Currently only gitea and docker are supported as remote and executor
respectively.
"""
from enum import Enum
from typing import NamedTuple, List


class TriggerFailed(Exception):
    "Failure to trigger a job"


class RemoteApiFailed(Exception):
    "Failure while working with a remote"


class JobStatus(NamedTuple):
    is_running: bool
    exit_code: int
    logs: str
    started_at: str
    finished_at: str


class Status(Enum):
    "Each pipeline can ONLY be in any one of these statuses"
    PENDING = 10
    RUNNING = 30
    FAILED = 40
    PASSED = 50
    TIMEOUT = 60
    SKIPPED = 70


class Repo:
    """
    Contains information about the current VCS repo.
    """

    def __init__(self, sha: str, branch: str, remote: str, commit_message: str):
        self.sha: str = sha
        self.branch: str = branch
        self.remote: str = remote
        self.commit_message: str = commit_message

    def files_changed(self, target: str) -> List[str]:
        """
        Returns list of file paths that have changed between current sha and
        target.
        """
        raise NotImplementedError()

    @classmethod
    def from_env(cls) -> "Repo":
        """
        Creates a :class:`~jaypore_ci.interfaces.Repo` instance
        from the environment and git repo on disk.
        """
        raise NotImplementedError()


class Executor:
    """
    An executor is something used to run a job.
    It could be docker / podman / shell etc.

    It must define `__enter__` and `__exit__` so that it can be used as a context manager.

    """

    def run(self, job: "Job") -> str:
        "Run a job and return it's ID"
        raise NotImplementedError()

    def __init__(self):
        self.pipe_id = None
        self.pipeline = None

    def set_pipeline(self, pipeline: "Pipeline") -> None:
        """Set the current pipeline to the given one."""
        self.pipe_id = id(pipeline)
        self.pipeline = pipeline

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        On exit the executor must clean up any pending / stuck / zombie jobs that are still there.
        """

    def get_status(self, run_id: str) -> JobStatus:
        """
        Returns the status of a given run.
        """
        raise NotImplementedError()


class Remote:
    """
    Something that allows us to show other people the status of the CI job.
    It could be gitea / github / gitlab / email system.

    Must define `__enter__` and `__exit__` so that it can be used as a context
    manager.
    """

    def __init__(self, *, sha, branch):
        self.sha = sha
        self.branch = branch

    def publish(self, report: str, status: str):
        """
        Publish this report somewhere.
        """
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @classmethod
    def from_env(cls, *, repo: "Repo"):
        """
        This function should create a Remote instance from the given environment.
        It can read git information / look at environment variables etc.
        """
        raise NotImplementedError()


class Reporter:
    """
    Something that generates the status of a pipeline.

    It can be used to generate reports in markdown, plaintext, html, pdf etc.
    """

    def render(self, pipeline: "Pipeline") -> str:
        """
        Render a report for the pipeline.
        """
        raise NotImplementedError()



import subprocess
from typing import List

from jaypore_ci.interfaces import Repo


class Git(Repo):
    def files_changed(self, target: str) -> List[str]:
        "Returns list of files changed between current sha and target"
        return (
            subprocess.check_output(
                f"git diff --name-only {target} {self.sha}", shell=True
            )
            .decode()
            .strip()
            .split("\n")
        )

    @classmethod
    def from_env(cls) -> "Git":
        """
        Gets repo status from the environment and git repo on disk.
        """
        remote = (
            subprocess.check_output(
                "git remote -v | grep push | grep https | awk '{print $2}'", shell=True
            )
            .decode()
            .strip()
        )
        assert "https://" in remote, f"Only https remotes supported: {remote}"
        assert ".git" in remote
        # NOTE: Later on perhaps we should support non-https remotes as well
        # since JCI does not actually do anything with the remote.
        branch = (
            subprocess.check_output(
                r"git branch | grep \* | awk '{print $2}'", shell=True
            )
            .decode()
            .strip()
        )
        sha = subprocess.check_output("git rev-parse HEAD", shell=True).decode().strip()
        message = (
            subprocess.check_output("git log -1 --pretty=%B", shell=True)
            .decode()
            .strip()
        )
        return cls(sha=sha, branch=branch, remote=remote, commit_message=message)



"""
The basic logging module.
"""
import logging
from typing import Any

import structlog

# This is used to accumulate logs and is later sent over to the CI status as a
# separate log list
jaypore_logs = []


class JayporeLogger:
    """
    This is mainly used to collect logs into a single global variable so that
    the logs of the CI runner itself can also be posted as part of the CI
    report.
    """

    def __getstate__(self) -> str:
        return "stdout"

    def __setstate__(self, state: Any) -> None:
        pass

    def __deepcopy__(self, memodict: dict[Any, Any] = None) -> "JayporeLogger":
        return self.__class__()

    def msg(self, message: str) -> None:
        global jaypore_logs  # pylint: disable=global-statement
        jaypore_logs.append(message)
        if len(jaypore_logs) > 1500:
            jaypore_logs = jaypore_logs[-1000:]
        print(message)

    log = debug = info = warn = warning = msg
    fatal = failure = err = error = critical = exception = msg


class JayporeLoggerFactory:
    def __init__(self):
        pass

    def __call__(self, *args) -> JayporeLogger:
        return JayporeLogger()


structlog.configure(
    processors=[
        structlog.contextvars.merge_contextvars,
        structlog.processors.add_log_level,
        # structlog.processors.StackInfoRenderer(),
        # structlog.dev.set_exc_info,
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.dev.ConsoleRenderer(colors=False),
    ],
    wrapper_class=structlog.make_filtering_bound_logger(logging.NOTSET),
    context_class=dict,
    logger_factory=JayporeLoggerFactory(),
    cache_logger_on_first_use=False,
)
logger = structlog.get_logger()
